import math
import random

import geopandas
import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame, GeoSeries
from pyproj import Transformer
from rasterio.features import rasterize
from shapely import Geometry, affinity


def load_geometries_from_paths(
    file_paths: list[str], target_crs: str | int = 4326
) -> GeoSeries:
    """Loads geometries from given paths into same CRS and returns them as a GeoSeries

    Args:
        file_paths (list[str]): paths to vector files containing geometries to load
        target_crs (str | int, optional): destination Coordinate Reference System (CRS). Defaults to 4326 (WGS).

    Returns:
        geopandas.GeoSeries: loaded geometries
    """
    valid_zones = []
    for path in file_paths:
        valid_zones.append(
            geopandas.read_file(path).make_valid().geometry.to_crs(target_crs)
        )
    valid_zones: geopandas.GeoSeries = pd.concat(valid_zones)
    return valid_zones


def geometries_to_mask(
    geometries: list[Geometry],
    affine_transform: list[int],
    out_shape: tuple[int],
    mask_value: int = 1,
) -> np.ndarray:
    """Transform georeferenced geometries into a mask with geometries filled in. Simple wrapper around
    rasterio.features.rasterize, see https://rasterio.readthedocs.io/en/stable/api/rasterio.features.html for
    more information.

    Args:
        geometries (list[Geometry]): list of geometries in geo-coordinates
        affine_transform (list[int]): transform from geo-coordinates into image coordinates
        out_shape (tuple[int]): shape of output mask
        mask_value (int, optional): value to burn into geometries. Defaults to 1.

    Returns:
        np.ndarray: generated mask
    """
    if len(geometries) == 0:
        return np.zeros(out_shape, dtype=np.uint8)
    rasterized_geoms = rasterize(
        shapes=geometries, out_shape=out_shape, transform=affine_transform
    )
    rasterized_geoms = (rasterized_geoms * mask_value).astype(np.uint8)
    return rasterized_geoms


def randomly_translate_geometries(
    geometries: GeoSeries,
    x_off_range: tuple[int],
    y_off_range: tuple[int],
    random_seed: int = None,
) -> GeoSeries:
    """Applies random translation to x and y coordinates for each geometry according to given translation ranges.

    Args:
        geoseries (GeoSeries): geometries to translate
        x_off_range (tuple[int]): translation range for X coordinate
        y_off_range (tuple[int]): translation range for Y coordinate
        random_seed (int, optional): set it to have reproducible results. Defaults to None.

    Returns:
        GeoSeries: translated geometries
    """
    if random_seed is not None:
        random.seed(random_seed)
    geometries = geometries.apply(
        lambda geom: affinity.translate(
            geom, xoff=random.randint(*x_off_range), yoff=random.randint(*y_off_range)
        )
    )
    return geometries


def transform_pt_crs(
    x: float, y: float, src_crs: int | str, dst_crs: int | str
) -> tuple[float, float]:
    """Transforms given point coordinates from one crs to another

    Args:
        x (float): horizontal (or east-west) coordinate
        y (float): vertical coordinate (or north-south)
        src_crs (int | str): source CRS
        dst_crs (int | str): destination CRS

    Returns:
        tuple[float, float]: transformed coordinates
    """
    transformer = Transformer.from_crs(src_crs, dst_crs)
    transformed_pt = transformer.transform(x, y)
    return transformed_pt


def buffer_wgs84_points(
    points: GeoDataFrame,
    tile_size_m: int,
    cap_style: int = 3,
    drop_epsg_col: bool = True,
) -> GeoDataFrame:
    """Given a geodataframe with point geometries in WGS84, buffers them and returns the resulting tiles in WGS84

    Args:
        spots (GeoDataFrame): geodataframe with point geometries
        tile_size_m (int): tiles will be of size (tile_size_m X tile_size_m)
        cap_style (int): the shape of the ends of the buffer. Defaults to 3. See
        https://shapely.readthedocs.io/en/stable/reference/shapely.buffer.html for
        more info
        drop_epsg_col (bool): whether to drop the epsg column generated for buffering the
        geometries

    Returns:
        GeoDataFrame: geodataframe with tiles
    """
    tile_centroids = points.copy()
    tile_centroids["epsg"] = tile_centroids.geometry.apply(
        lambda pt: find_best_utm_code(pt.x, pt.y)
    )
    tile_centroids_grouped = tile_centroids.groupby("epsg")
    tiles = []
    for epsg, epsg_hotspots in tile_centroids_grouped:
        epsg_hotspots = epsg_hotspots.to_crs(epsg=epsg)
        group_tiles = epsg_hotspots.geometry.buffer(
            tile_size_m // 2, cap_style=cap_style
        )
        group_tiles = group_tiles.to_crs(epsg=4326)
        if not drop_epsg_col:
            epsg_hotspots["epsg"] = epsg
        group_tiles = GeoDataFrame(
            epsg_hotspots,
            geometry=group_tiles,
            crs=4326,
        )
        tiles.append(group_tiles)
    tiles = pd.concat(tiles)
    return tiles


def create_cell_grid(
    bounds: tuple[float, float, float, float],
    n_cells_x: int,
    n_cells_y: int,
    out_crs: str = "EPSG:4326",
) -> geopandas.GeoDataFrame:
    """Create a cell grid of size (n_cells_x,n_cells_y) within given bounds.

    Args:
        bounds (tuple[float, float, float, float]): bounds in which to create the grid in the format west, north, east, south
        n_cells_x (int): number of cells in x direction
        n_cells_y (int): number of cells in y direction
        out_crs (_type_, optional): Output Coordinate Reference System. Defaults to "EPSG:4326".

    Returns:
        geopandas.GeoDataFrame: grid cell as a geodataframe
    """
    xmin, ymin, xmax, ymax = bounds
    cell_size_x = (xmax - xmin) / n_cells_x
    cell_size_y = (ymax - ymin) / n_cells_y
    grid_cells = []
    for x0 in np.arange(xmin, xmax, cell_size_x):
        for y0 in np.arange(ymin, ymax, cell_size_y):
            x1 = x0 - cell_size_x
            y1 = y0 + cell_size_y
            grid_cells.append(shapely.geometry.box(x0, y0, x1, y1))
    return geopandas.GeoDataFrame(grid_cells, columns=["geometry"], crs=out_crs)


def find_best_utm_code(lon: float, lat: float) -> str:
    """finds the best matching UTM code for given lat-long coordinates

    Args:
        lon (float): longitude
        lat (float): ñatitude

    Returns:
        str: UTM code found
    """
    utm_band = str((math.floor((lon + 180) / 6) % 60) + 1)
    if len(utm_band) == 1:
        utm_band = "0" + utm_band
    if lat >= 0:
        epsg_code = "326" + utm_band
    else:
        epsg_code = "327" + utm_band
    return epsg_code



from typing import Any, Literal

import affine
import numpy as np
import rasterio
from geopandas import GeoSeries
from rasterio.mask import mask
from rasterio.merge import merge
from rasterio.warp import Resampling, calculate_default_transform, reproject
from rio_cogeo import cog_translate
from shapely import Polygon, box


def compute_affine_transform(
    min_x: float, max_y: float, resolution_x: float, resolution_y: float
) -> affine.Affine:
    """Creates the affine transform for given coordinates and resolution"""
    return affine.Affine(
        resolution_x,
        0,
        min_x,
        0,
        resolution_y,
        max_y,
    )


def get_raster_metadata(input_path: str) -> dict[str, Any]:
    """Retrieves metadata from raster stored in disk"""
    with rasterio.open(input_path) as src:
        return src.meta


def get_raster_extent(input_path: str) -> GeoSeries:
    """Retrieves the extent (bounding box) of the raster as GeoSeries in the same CRS as the raster"""
    with rasterio.open(input_path) as src:
        geom = box(*src.bounds)
        return GeoSeries.from_wkt(data=[geom.wkt], crs=src.crs)


def get_raster_footprint(input_path: str) -> GeoSeries:
    """Retrieves the footprint of the raster file as a GeoSeries of polygons in WGS84

    Args:
        input_path (str): path to raster from which to extract footprint

    Returns:
        GeoSeries: footprint of raster
    """
    with rasterio.open(input_path) as src:
        # read dataset valid mask
        mask = src.dataset_mask()

        # Extract feature shapes and values from the array.
        footprint = []
        for geom, val in rasterio.features.shapes(mask, transform=src.transform):
            # skip masks that are not valid
            if val != 255:
                continue

            # Transform shapes from the dataset's own coordinate
            # reference system to CRS84 (EPSG:4326).
            geom = rasterio.warp.transform_geom(src.crs, "EPSG:4326", geom, precision=6)
            for g in geom["coordinates"]:
                footprint.append(Polygon(g))
        return GeoSeries(footprint, crs="EPSG:4326")


def mask_raster_pixels_by_geometries(
    geometries: GeoSeries,
    raster_path: str,
    output_path: str,
    invert: bool = False,
    fill_value: int = 0,
):
    """Masks raster according to given geometries. Pixel values outside of geometries will be replaced
    by given `fill_value`, unless `inverted=True` in which case it's the opposite. This function is a wrapper
    around rasterio.mask.mask

    Args:
        geometries (geopandas.GeoSeries): masking geometries
        raster_path (str): path to input raster
        output_path (str): path to output masked raster
        invert (bool, optional): set to True if pixels inside geometries should be replaced. Defaults to False.
        fill_value (int, optional): value to replace. Defaults to 0.
    """
    with rasterio.open(raster_path) as src:
        profile = src.profile
        prj_geometries = geometries.to_crs(src.crs)
        masked_array, out_transform = mask(
            src,
            prj_geometries,
            invert=invert,
            nodata=fill_value,
        )
        profile.update(
            {
                "height": masked_array.shape[1],
                "width": masked_array.shape[2],
                "transform": out_transform,
            }
        )
    with rasterio.open(output_path, "w", **profile) as dst:
        dst.write(masked_array)


def merge_rasters_into_mosaic(
    raster_paths: list[str],
    mosaic_path: str,
    method: Literal["first", "last", "min", "max"] = "max",
):
    """Merges a list of rasters into a single mosaic. Assumes all input rasters are in the same CRS
    This function is a simple wrapper around rasterio.merge.merge, see https://rasterio.readthedocs.io/en/stable/api/rasterio.merge.html
    for more complex functionalities.

    Args:
        paths (list[str]): list of rasters to merge
        mosaic_path (str): output mosaic path
        method (str): which method to use for merging. Defaults to "max"
    """
    mosaic, output_trf = merge(raster_paths, method=method)
    with rasterio.open(raster_paths[0]) as ref_src:
        ref_meta = ref_src.meta
    output_meta = ref_meta.copy()
    output_meta.update(
        {
            "driver": "GTiff",
            "height": mosaic.shape[1],
            "width": mosaic.shape[2],
            "transform": output_trf,
        }
    )
    with rasterio.open(mosaic_path, "w", **output_meta) as dst:
        dst.write(mosaic)


def transform_to_cog(raster_path: str, output_path: str):
    """Translates a GeoTIFF raster to Cloud Optimized GeoTIFF (COG) format. It is a simple wrapper around
    cog_translate function from rio-cogeo library. Use https://cogeotiff.github.io/rio-cogeo/ if you need more
    control.


    Args:
        raster_path (str): path to raster to transform
        output_path (str): path to cog destination
    """
    with rasterio.open(raster_path) as src:
        src_profile = src.profile
    src_profile.update(
        {
            "GTIFF": "IF_SAFER",
            "interleave": "pixel",
            "compress": "DEFLATE",
            "tiled": True,
            "blockxsize": 256,
            "blockysize": 256,
        }
    )

    config = dict(
        GDAL_NUM_THREADS="ALL_CPUS",
        GDAL_TIFF_INTERNAL_MASK=True,
        GDAL_TIFF_OVR_BLOCKSIZE="128",
    )
    cog_translate(raster_path, output_path, src_profile, config=config, in_memory=False)


def clip_raster_borders(
    raster_path: str, boundary_cut_m: int, nodata: Any = 0
) -> tuple[np.ndarray, dict[str, Any]]:
    """Clips borders of a raster by given distance in meters. Returns the clipped array and georeference metadata

    Args:
        raster_path (str): path to raster to clip
        boundary_cut_m (int): clipping distance from border, in meters

    Returns:
        tuple[np.ndarray, dict[str, Any]]: clipped raster array and metadata
    """
    with rasterio.open(raster_path) as src:
        bounds = src.bounds
        clipping_geom = GeoSeries(
            Polygon(
                [
                    (bounds.left, bounds.top),
                    (bounds.right, bounds.top),
                    (bounds.right, bounds.bottom),
                    (bounds.left, bounds.bottom),
                ]
            ),
            crs=src.crs,
        )
        clipping_geom = clipping_geom.buffer(distance=-boundary_cut_m)
    clipped_raster, meta = clip_raster_by_geom(
        raster_path, clipping_geom, nodata=nodata
    )
    return clipped_raster, meta


def clip_raster_by_geom(
    raster_path: str, geom: GeoSeries, nodata: int = 0, all_touched: bool = True
) -> tuple[np.ndarray, dict[str, Any]]:
    """Clips a raster by given geometry. Returns the clipped array and georeference metadata. Uses rasterio.mask.mask
    underneath

    Args:
        raster_path (str): path to raster
        geom (GeoDataFrame): clipping geometry
        nodata (int, optional): value for nodata in raster. Defaults to 0.
        all_touched (bool, optional): whether to consider all pixels touched by the geometry. Defaults to True.

    Returns:
        tuple[np.ndarray, dict[str, Any]]: clipped array and raster metadata
    """
    with rasterio.open(raster_path) as src:
        # clipping geometry must have the same CRS as raster
        geometry_src_crs = geom.to_crs(src.crs)
        out, out_transform = mask(
            src,
            shapes=geometry_src_crs.geometry,
            crop=True,
            all_touched=all_touched,
            nodata=nodata,
        )
        # update metadata
        out_meta = src.meta
        out_meta.update(
            {
                "driver": "GTiff",
                "height": out.shape[1],
                "width": out.shape[2],
                "transform": out_transform,
                "nodata": nodata,
            }
        )
    return out, out_meta


def mask_file_to_vector_file(
    mask_path: str,
    output_path: str,
    mask_value: int = 1,
    simplify: bool = False,
    simplify_tol: float = 1.0,
):
    """Converts a raster file containing an integer mask to a vector file with constituent polygons.
    The projection of the polygons will match that of the input raster. User can optionally simplify the
    resulting geometries via the Douglas Peucker algorithm.

    Args:
        mask_path (str): path to raster from where polygons will be extracted
        output_path (str): file path to save polygons as vector file
        mask_value (int, optional): the method will look for pixel values equal to this parameter
        for creating the geometries. Defaults to 1.
        simplify (bool, optional): whether to simplify the geometries. Defaults to False.
        simplify_tol (float, optional): tolerance (in raster projection units) for simplification. Defaults to 1.0.
    """
    with rasterio.open(mask_path) as src:
        mask = src.read()
        mask[mask != mask_value] = 0
        mask[mask == mask_value] = 1
        crs = src.crs
        transform = src.transform
    polygons = mask_to_polygons(mask, crs, transform)
    if simplify:
        polygons = polygons.simplify(simplify_tol)
    polygons.to_file(output_path)


def mask_to_polygons(mask: np.ndarray, crs: str, transform: np.ndarray) -> GeoSeries:
    """Converts a binary mask into a GeoSeries of polygons.

    Args:
        mask (np.ndarray): binary mask
        crs (str): coordinate reference system of the mask
        transform (np.ndarray): geo transform of the mask

    Returns:
        GeoSeries: polygons extracted from the mask
    """
    polygons = []
    for geom, val in rasterio.features.shapes(mask, transform=transform):
        if val == 0:
            continue
        for g in geom["coordinates"]:
            polygons.append(Polygon(g))
    polygons = GeoSeries(polygons, crs=crs)
    return polygons


def reproject_raster(
    raster_path: str,
    output_path: str,
    dst_crs: str,
    float_output: bool = True,
    resampling_method: Resampling = Resampling.nearest,
):
    """Reprojects a raster into given CRS

    Args:
        raster_path (str): input raster path
        output_path (str): output reprojected raster path
        dst_crs (str): destination coordinate reference system
        float_output (bool, optional): transform output raster to float and use NaN as nodata for output raster.
        Rasterio reprojection doesn't handle nodata well, so it is recommended to keep this value true. Defaults to True.
        resampling_method (Resampling, optional): resampling method for interpolating reprojected raster pixels. Defaults to Resampling.nearest.
    """
    with rasterio.open(raster_path) as src:
        # calculate the transform matrix for the output
        dst_transform, width, height = calculate_default_transform(
            src.crs,
            dst_crs,
            src.width,
            src.height,
            *src.bounds,  # unpacks outer boundaries (left, bottom, right, top)
        )
        # set properties for output
        dst_kwargs = src.meta.copy()
        dst_kwargs.update(
            {
                "crs": dst_crs,
                "transform": dst_transform,
                "width": width,
                "height": height,
            }
        )
        if float_output:
            dst_kwargs.update({"nodata": np.nan, "dtype": np.float32})

        with rasterio.open(output_path, "w", **dst_kwargs) as dst:
            # iterate through bands
            for i in range(1, src.count + 1):
                reproject(
                    source=rasterio.band(src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src.crs,
                    dst_transform=dst_transform,
                    dst_crs=dst_crs,
                    resampling=resampling_method,
                )



# -*- coding: utf-8 -*-
"""

    Python logging tuned to extreme.

"""

__author__ = "Mikko Ohtamaa <mikko@opensourcehacker.com>"
__license__ = "MIT"

import logging

from logutils.colorize import ColorizingStreamHandler


class RainbowLoggingHandler(ColorizingStreamHandler):
    """A colorful logging handler optimized for terminal debugging aestetichs.

    - Designed for diagnosis and debug mode output - not for disk logs

    - Highlight the content of logging message in more readable manner

    - Show function and line, so you can trace where your logging messages
      are coming from

    - Keep timestamp compact

    - Extra module/function output for traceability

    The class provide few options as member variables you
    would might want to customize after instiating the handler.
    """

    # Define color for message payload
    level_map = {
        logging.DEBUG: (None, "cyan", False),
        logging.INFO: (None, "white", False),
        logging.WARNING: (None, "yellow", True),
        logging.ERROR: (None, "red", True),
        logging.CRITICAL: ("red", "white", True),
    }

    date_format = "%H:%M:%S"

    #: How many characters reserve to function name logging
    who_padding = 22

    #: Show logger name
    show_name = True

    def get_color(self, fg=None, bg=None, bold=False):
        """
        Construct a terminal color code

        :param fg: Symbolic name of foreground color

        :param bg: Symbolic name of background color

        :param bold: Brightness bit
        """
        params = []
        if bg in self.color_map:
            params.append(str(self.color_map[bg] + 40))
        if fg in self.color_map:
            params.append(str(self.color_map[fg] + 30))
        if bold:
            params.append("1")

        color_code = "".join((self.csi, ";".join(params), "m"))

        return color_code

    def colorize(self, record):
        """
        Get a special format string with ASCII color codes.
        """

        # Dynamic message color based on logging level
        if record.levelno in self.level_map:
            fg, bg, bold = self.level_map[record.levelno]
        else:
            # Defaults
            bg = None
            fg = "white"
            bold = False

        # Magician's hat
        # https://www.youtube.com/watch?v=1HRa4X07jdE
        template = [
            "[",
            self.get_color("cyan", None, True),
            "%(asctime)s",
            self.reset,
            "] ",
            self.get_color("white", None, True) if self.show_name else "",
            "%(name)s " if self.show_name else "",
            "%(padded_who)s",
            self.reset,
            " ",
            self.get_color(bg, fg, bold),
            "%(message)s",
            self.reset,
        ]

        format = "".join(template)

        who = [
            self.get_color("green"),
            getattr(record, "funcName", ""),
            "()",
            self.get_color("black", None, True),
            ":",
            self.get_color("cyan"),
            str(getattr(record, "lineno", 0)),
        ]

        who = "".join(who)

        # We need to calculate padding length manualy
        # as color codes mess up string length based calcs
        unformatted_who = (
            getattr(record, "funcName", "")
            + "()"
            + ":"
            + str(getattr(record, "lineno", 0))
        )

        if len(unformatted_who) < self.who_padding:
            spaces = " " * (self.who_padding - len(unformatted_who))
        else:
            spaces = ""

        record.padded_who = who + spaces

        formatter = logging.Formatter(format, self.date_format)
        self.colorize_traceback(formatter, record)
        output = formatter.format(record)
        # Clean cache so the color codes of traceback don't leak to other formatters
        record.ext_text = None
        return output

    def colorize_traceback(self, formatter, record):
        """
        Turn traceback text to red.
        """
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            record.exc_text = "".join(
                [
                    self.get_color("red"),
                    formatter.formatException(record.exc_info),
                    self.reset,
                ]
            )

    def format(self, record):
        """
        Formats a record for output.

        Takes a custom formatting path on a terminal.
        """
        if self.is_tty:
            message = self.colorize(record)
        else:
            message = logging.StreamHandler.format(self, record)

        return message


def create_rainbow_logger(name: str) -> logging.Logger:
    import sys

    logger = logging.getLogger(name)

    handler = RainbowLoggingHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


if __name__ == "__main__":
    # Run test output on stdout
    import sys

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    handler = RainbowLoggingHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)
    logger = logging.getLogger("test")

    def test_func():
        logger.debug("debug msg")
        logger.info("info msg")
        logger.warn("warn msg")

    def test_func_2():
        logger.error("error msg")
        logger.critical("critical msg")

        try:
            raise RuntimeError("Opa!")
        except Exception as e:
            logger.exception(e)

    test_func()
    test_func_2()



import json
import logging
import os
from logging import Logger
from typing import Any

import matplotlib.pyplot as plt
import mlflow
from pydantic import BaseModel

from experiment_goodies.logger.rainbow_logger import RainbowLoggingHandler


class ExperimentLoggerConfig(BaseModel):
    log_level: str = "INFO"
    log_freq_steps: int = 100
    log_mlflow: bool = False


class ExperimentLogger(Logger):
    """Extends Logger from logging library to incorporate functionalities useful during model training.

    Attributes:
        name (str): owner of this logger (will be printed to console)
        experiment_config (ExperimentConfig): specifies experiment configuration such as network values and dataset structure
        exp_folder (str, optional): folder in which to store logging artifacts generated in this experiment. Defaults to "".

    For mlflow logging purposes, it is assumed that there is a .env file in working directory containing tracking variables:
        MLFLOW_TRACKING_URI (e.g. http://localhost:5000)
        MLFLOW_TRACKING_USERNAME (e.g. admin)
        MLFLOW_TRACKING_PASSWORD (e.g. password)
    """

    def __init__(
        self,
        name: str,
        logger_config: ExperimentLoggerConfig,
        exp_folder: str = "",
    ):
        super().__init__(name, logger_config.log_level)
        self.logger_config = logger_config
        handler = RainbowLoggingHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        self.addHandler(handler)
        self.exp_folder = exp_folder

    def __save_exp_config(self, config: dict[str, Any]):
        """Saves the configuration of this experiment"""
        with open(os.path.join(self.exp_folder, "experiment_config.json"), "w") as f:
            json.dump(config, f, indent=4)

    def log_config_to_mlflow(
        self, experiment_config: dict[str, Any], dvc_file_path: str
    ) -> mlflow.ActiveRun:
        """Logs parameters and dvc file to mlflow"""
        mlflow.log_params(experiment_config)
        mlflow.log_artifact(dvc_file_path)

    def log_metrics(self, metrics: dict, epoch: int, step: int, total_step: int):
        """Logs training metrics

        Args:
            metrics (dict): dictionary containing metrics. (eg. {"val_loss":0.509})
            epoch (int): training epoch
            step (int): training step at which metrics were computed
            total_step (int): total training steps in this experiment
        """
        step_info = f"{step}/{total_step}"
        log_msg = f"Epoch: {epoch} | step: {step_info} | "
        for k, v in metrics.items():
            log_msg += f"{k}: {v:.4f} | "
        self.info(log_msg)
        if self.logger_config.log_mlflow:
            mlflow.log_metrics(metrics, step=step)

    def log_epoch_info(self, epoch: int, step: int):
        if self.logger_config.log_mlflow:
            mlflow.log_metric("epoch", epoch, step=step)

    def log_prediction_figure(self, figure: plt.Figure, epoch: int, fig_id: int):
        """Saves a sample prediction figure to disk and logs it to mlflow

        Args:
            figure (plt.Figure): figure containing sample prediction
            epoch (int): epoch of training
            fig_id (int): figure identifier
        """
        fig_path = os.path.join(
            self.exp_folder,
            "figures",
            f"epoch{epoch}_pred_{fig_id}.png",
        )
        figure.savefig(fig_path)
        if self.logger_config.log_mlflow:
            mlflow.log_figure(
                figure, f"predictions_by_epoch/{epoch}/prediction_{fig_id}.png"
            )
        plt.close()

    def log_experiment_config(self, exp_config: dict[str, Any], dvc_file_path: str):
        """Saves configuration, or parameters, of this experiment

        Args:
            exp_config (dict[str, Any]): experiment configuration to be logged
            dvc_file_path (str): path to dvc file of experiment dataset
        """
        self.__save_exp_config(exp_config)
        if self.logger_config.log_mlflow:
            self.log_config_to_mlflow(exp_config, dvc_file_path)
